//  Dec 1st 2003

// LevelMeasure.c   from LevelMeasureEngine.c

/******************************************************************************/
/*	LevelMeasureEngine.c - Level Measurement Engine
/*
/*	Copyright (c) 2000, Product Technology Partners Ltd
/*
/*	Author	Martin Saxon
/******************************************************************************/

/*	Purpose
/*	-------
/*
/*                                                                            */

/*	History
/*	-------
/*	24-Jul-2000	MJS	Created
/*	25-Aug-2000	JTW	Fixed problem with using _inline with function sqr
/*                                                                            */

/******************************************************************************/


// BG Nov 2002
// added get_levels((char* sigfile, FILE *sig_fd, ULONG *total_bytes, ULONG *sig_len_bytes, struct formathead *sig_head,
//									 double *RMS, double *P56, double *PEAK)
// for measuring mono & stereo wav file levels in audibility tool

// Calibration note
// Originally this did levels wrt 'the level from a square wave would be 0 dB'
// As a calibration level we use the level of a full scale sinusoid, e.g = 100 dB SPL
// RMS of a sinusoid = sqrt(2)/2 = 0.707 of the peak value = 20log10(sqrt(2)/2) = -3 dB vs square wave
// Therefore the level of a signal vs a full scale sinusoid is 3 dB higher than the level vs full scale square wave


#if 0
commented out test main

short verbose;

void main()
{
void get_levels(char* sigfile, FILE *sig_fd, ULONG *total_bytes, ULONG *sig_len_bytes, struct formathead *sig_head,
								double *RMS, double *P56, double *PEAK);

//	char sigfile[] = {"1k90.wav"};
//	char sigfile[] = {"1k.wav"};
//	char sigfile[] = {"chimes.wav"};
	char sigfile[] = {"ctmelody.wav"};

	double RMS[2], P56[2], PEAK[2]; // [0] for mono files or left channel of stereo, [1] for right channel of stereo wav files
	ULONG total_bytes;
	ULONG sig_len_bytes;
	struct formathead sig_head;
	FILE *sig_fd = NULL;

	printf("\nfile: %s\n\n",sigfile);

	get_levels(sigfile, sig_fd, &total_bytes, &sig_len_bytes, &sig_head, RMS, P56, PEAK);

	if(sig_head.nchans == 2)
		printf("\nStereo left  channel:\n");
	else printf("\nMono:\n");
	printf("InstPeakLevelMeasure   %.3f\n",PEAK[0]);
	printf("RMSLevelMeasure        %.3f\n",RMS[0]);
	printf("P56LevelMeasure        %.3f\n",P56[0]);

	if(sig_head.nchans == 2){
		printf("\nStereo right channel:\n");
		printf("InstPeakLevelMeasure   %.3f\n",PEAK[1]);
		printf("RMSLevelMeasure        %.3f\n",RMS[1]);
		printf("P56LevelMeasure        %.3f\n",P56[1]);
	}
}
//------------------------------------------

#endif




#include "readfile.h" // need to be before LevelMeasureEngine.h for eg ULONG definition
#include "LevelMeasureEngine.h"


#define NLEVELS 40
#define LEVELFACTOR 0.5
#define NOMINALSAMPLERATE 32000 // 0
#define SMOOTHINGTC 0.03
#define HANGOVERTIME 0.2
#define MARGIN 15.9


extern struct formathead{
	USHORT w_format;    /* 16 bit number = 1 for PCM */
	USHORT nchans;      /* 1 or 2 channels */
	ULONG  clockHz;     /* sampling clock */
	ULONG  bytes_sec;   /* clockHz*nchans*bytes_sample */
	USHORT bytes;       /* 2=16 bit mono, 4 = 16 bit stereo */
	USHORT bits_sample; /* 16 for us */
	ULONG  data_bytes;  /* from DATA chunk */
};

//#ifdef WIN32 /* The use of _inline is Microsoft specific! */
//static double _inline sqr(double a)
//#else
static double sqr(double a)
//#endif
{
	return a*a;
}

double P56LevelMeasure(double *inSamples, int nInSamples, int sampleRate, double *longTermLevel, double *activityFactor)
{
	double c[NLEVELS];
	int h[NLEVELS];
	int a[NLEVELS];
	double p=0.0;
	double q=0.0;
	double s=0.0;
	double sq=0.0;
	double maxval=0.0;
	double g, g1;
	int i;
	double totalTime;
	double longTermPower;
	double activeLevelEstimate;
	double ale[NLEVELS];
	double tlevel[NLEVELS];
	int sampleSkip;
	double skippedRate;
	int nSamplesConsidered = 0;

	int indx;

	// Calculate time constants, etc.
	if (0==NOMINALSAMPLERATE)
	{
		sampleSkip = 1;
		skippedRate = sampleRate;
	}
	else
	{
		sampleSkip = sampleRate / NOMINALSAMPLERATE;
		if (sampleSkip<=0) sampleSkip = 1;
		skippedRate = ((double)sampleRate)/sampleSkip;
	}

	g = exp(-1.0/(skippedRate*SMOOTHINGTC));
	g1 = 1.0-g;
	i = (int)ceil(HANGOVERTIME*skippedRate);

	// We first need to calculate some working data
	for (indx=0;indx<nInSamples;indx+=sampleSkip)
	{
		double thisVal = inSamples[indx];
		s += thisVal;
		sq += sqr(thisVal);
		thisVal=fabs(thisVal);
		if (thisVal>maxval) maxval = thisVal;
		nSamplesConsidered++;
	}

	if (sq<=0.0)
		return -99999.0; // No samples or all zero - small level!

	c[0]=maxval; a[0]=0; h[0]=i;
	for (indx=1;indx<NLEVELS;indx++)
	{
		c[indx]=LEVELFACTOR*c[indx-1];
		a[indx]=0;
		h[indx]=i;
	}


	// Now do the necessary calculations
	for (indx=0;indx<nInSamples;indx+=sampleSkip)
	{
		int jndx;

		p = g*p + g1*fabs(inSamples[indx]);
		q = g*q + g1*p;
		
		for (jndx=0;jndx<NLEVELS;jndx++)
			if (q>=c[jndx])
			{
				a[jndx]++;
				h[jndx]=0;
			}
			else
			{
				if (h[jndx]<i)
				{
					a[jndx]++;
					h[jndx]++;
				}
			}
	}

	// Now we can work out the statistics
	totalTime = ((double)nSamplesConsidered)/skippedRate;
	longTermPower = sq/nSamplesConsidered;
	*longTermLevel = 10.0*log10(longTermPower);

	activeLevelEstimate = -99999.0;

	for (indx=0;indx<NLEVELS;indx++)
	{
		if (a[indx]>0)
			ale[indx] = 10.0*log10(sq/a[indx]);
		else
			ale[indx] = -99999.0;
		tlevel[indx] = ale[indx] - 20.0*log10(c[indx]);

		if (indx>0)
		{
			if ((tlevel[indx]>=MARGIN) && (tlevel[indx-1]<MARGIN))
			{
				// Interpolate to find right level
				double alpha = (tlevel[indx]-MARGIN)/(tlevel[indx]-tlevel[indx-1]);
				double alpha1 = 1.0-alpha;

				activeLevelEstimate = alpha*ale[indx-1]+alpha1*ale[indx];
				break;
			}
		}
	}

	if (activeLevelEstimate < -99998.0)
		*activityFactor = 0.0;
	else
		*activityFactor = 100.0*pow(10.0,0.1*(*longTermLevel-activeLevelEstimate));

	return activeLevelEstimate + 3.0103;  // see calibration note above
}


double RMSLevelMeasure(double *inSamples, int nInSamples)
{
	double accumulator=0.0;
	int indx;

	for (indx=0;indx<nInSamples;indx++)
		accumulator += sqr(inSamples[indx]);


//printf("My RMSLevelMeasure: rms %.1f\n",32768.*sqrt(accumulator/nInSamples));
//printf("If a full scale sinusoid = 0 dB, we would have %.1f dB\n",20*log10(sqrt(2.0)*sqrt(accumulator/nInSamples)));
	if ((accumulator<=0.0) || (nInSamples==0))
		return -99999.0;
	else
		return 10*log10(accumulator/nInSamples) + 3.0103; // see calibration note above
}

double InstPeakLevelMeasure(double *inSamples, int nInSamples)
{
	double maxVal = 0.0;
	int indx;

	for (indx=0;indx<nInSamples;indx++)
	{
		double aval = fabs(inSamples[indx]);
		if (aval>maxVal)
			maxVal = aval;
	}

//printf("InstPeakLevelMeasure:  maxVal = %.1f\n",maxVal*32768);
	if (maxVal<=0.0)
		return -99999.0;
	else
		return 20*log10(maxVal);
}

//--------------------------------------------------


void get_levels(char* sigfile, FILE *sig_fd, ULONG *total_bytes, ULONG *sig_len_bytes, struct formathead *sig_head,
								double *RMS, double *P56, double *PEAK)
{
	short *inData;
	double *dInData1, *dInData2;
	double longTermLevel;
	double activityFactor;
	int nSamples;
	int indx;
	int nChans;
	int sampleRate;

	read_headers(sigfile, &sig_fd, total_bytes, sig_len_bytes, sig_head); // file will be opened in read_headers but not closed

	nChans     = sig_head->nchans;  // 1 or 2
	nSamples   = sig_head->data_bytes/(nChans*2);
	sampleRate = sig_head->clockHz;
	
	(inData = (short *)malloc(sizeof(short)*nSamples*nChans)); // array for holding total waveform
	fread(inData,sizeof(short),nSamples*nChans,sig_fd); // read in whole file
	fclose(sig_fd); // close file opened in read_headers()

	if(nChans == 1){ // mono
		dInData1=(double *)malloc(sizeof(double)*nSamples);
		for (indx=0;indx<nSamples;indx++)
			dInData1[indx] = inData[indx] / 32768.0;

	} else if(nChans == 2) {
		dInData1=(double *)malloc(sizeof(double)*nSamples);
		dInData2=(double *)malloc(sizeof(double)*nSamples);
		for (indx=0;indx<nSamples;indx++){
			dInData1[indx] = inData[indx*2] / 32768.0;   // extract left channel
			dInData2[indx] = inData[indx*2+1] / 32768.0; // extract right channel
		}
	}

	PEAK[0] = InstPeakLevelMeasure(dInData1,nSamples);
	RMS[0]  = RMSLevelMeasure(dInData1,nSamples);
	P56[0]  = P56LevelMeasure(dInData1,nSamples, sampleRate, &longTermLevel, &activityFactor);


//	longTermLevel and activityFactor values calculatd but not returned
	
//	printf("InstPeakLevelMeasure   %.3f\n",InstPeakLevelMeasure(dInData1,nSamples));
//	printf("RMSLevelMeasure        %.3f\n",RMSLevelMeasure(dInData1,nSamples));
//	printf("P56LevelMeasure        %.3f\n",P56LevelMeasure(dInData1,nSamples, sampleRate, &longTermLevel, &activityFactor));
	if(nChans == 2){
		PEAK[1] = InstPeakLevelMeasure(dInData2,nSamples);
		RMS[1]  = RMSLevelMeasure(dInData2,nSamples);
		P56[1]  = P56LevelMeasure(dInData2,nSamples, sampleRate, &longTermLevel, &activityFactor);

//		printf("stereo right channel:\n");
//		printf("InstPeakLevelMeasure   %.3f\n",InstPeakLevelMeasure(dInData2,nSamples));
//		printf("RMSLevelMeasure        %.3f\n",RMSLevelMeasure(dInData2,nSamples));
//		printf("P56LevelMeasure        %.3f\n",P56LevelMeasure(dInData2,nSamples,sampleRate, &longTermLevel, &activityFactor));
	}	
}
