//  Dec 1st 2003

/* fir_sub1.c    from fir_wav2.c for 32k wav files from fir_corr.c   empties buffer to file at end    for stereo wav files */
/*
==============================================================
fir_wav2.c: C code to apply compensatory fir filter to audio.
Origin is time-shifted by half-aperture width...............
MAStone, 26-May-2001: for use in BRG recruitment simulation
===============================================================
*/

#include "readfile.h"

extern struct formathead{
	USHORT w_format;    /* 16 bit number = 1 for PCM */
	USHORT nchans;      /* 1 or 2 channels */
	ULONG  clockHz;     /* sampling clock */
	ULONG  bytes_sec;   /* clockHz*nchans*bytes_sample */
	USHORT bytes;       /* 2=16 bit mono, 4 = 16 bit stereo */
	USHORT bits_sample; /* 16 for us */
	ULONG  data_bytes;  /* from DATA chunk */
};

short verbose;


#define SGMNT 16384  /* how many samples to process */

#define MAX_HALF 2048	/* maximum halfwidth of fir filter */
#define FLOATS sizeof(float)
#define SIGMAX  32767.
#define SIGMIN -32767. /* helps with rectifying if symmetric */
#define NULLPTR (void *)0
#define UPDATE 0 /* (how many buffers-1) before update sample numbers to screen */ 
#define TRUE  1
#define FALSE 0


// GLOBAL
/* static data areas defined here */
float filter[2*MAX_HALF+1]; /* static coefficient space */
short in[2][SGMNT], out[2][SGMNT]; /* filter input/output */


/*--------------------------------------------------------------*/
void error(char *msg1, char *msg2){/* print user error and quit */
  fprintf(stderr,"\nERROR: Quitting fir_wav2 due to:\n\t%s %s\n",msg1,msg2);
  exit(1);
}
/*--------------------------------------------------------------*/
void zero_short_array(short *elemnt, short howmany, short nchans){
  /* Initialise an array of short to zero */
  register short i,j;
	for( j = 0 ; j < nchans; j++)
		for(i = howmany; i-- > 0; ) *(elemnt++) = 0;
}
/*-------------------------------------------*/
void zero_float_array(float *elemnt, short howmany){
  /* Initialise an array of float to zero */
  register short i;
  for(i = howmany; i-- > 0; ) *(elemnt++) = 0.0;
}
/*---------------------------------------------------------------*/
void convert_to_float(short *input, float *output, long howmany){
  /* convert input array to float */
  register short i;
  for(i = 0; i++ < howmany; ) *(output++) = (float) *(input++);
}
/*---------------------------------------------------------------*/
void convert_to_short(float *input, short *output, short howmany){
  /* convert input array to short */
  register short i;
  for(i = 0; i++ < howmany; ) *(output++) = (short) *(input++);
}
/*---------------------------------------------------------------------------------*/
short rd_fir_file(FILE *fir_file, float *coeffs, float gain){ /* read filter file */
  short idx, n_cfs;
  char sntnc[80];
  float tmp;
  
  fscanf(fir_file, "%[^\n]", &sntnc);
  
  fscanf(fir_file, "%hd", &n_cfs); /* second line has number of coefficients */
  
  if(verbose) printf("\nFIR filter [%s], %d coeffs",sntnc,n_cfs);
  if (n_cfs > (2*MAX_HALF+1)){
    sprintf(sntnc,"%d but can only handle %d coefficients",n_cfs,(2*MAX_HALF+1));
    error("filter file too big:",sntnc);
  }
  for (idx = 0; idx < n_cfs; idx ++){
    if(fscanf(fir_file, "%f", &tmp) == EOF) break;
    *(coeffs+idx) = tmp*gain;
  }
  if (idx != n_cfs) {
    sprintf(sntnc,"%d but only found %d coefficients",n_cfs,idx);
    error("filter file incorrect size:",sntnc);
  }
  if(verbose)printf("\nGain factor %6.3f (default = 0.997)\n",gain);
  return(n_cfs);
}
/*-----------------------------------------------------------------------------*/
float do_fir(float *in, float *fir_filt,short ntaps){ /* perform fir filter fn */
  float tmp;
  short idx;
  tmp = 0.;
  for (idx = 0; idx < ntaps; idx++) tmp += (*(in++) * *(fir_filt++));
  return(tmp);
}
/*-------------------------------------------------------*/



//void main(short argc, char **argv){




void fir_sub1(char* inputname, FILE **reading_from, char* temp_fir_file,  char* fir_name, struct formathead *fh)
{
  /* internal variables */
  short   ix, *shrt_ptr[2], fir_size, half_fir, first_buf;
	short		short_sig = FALSE; // TRUE for sig less than half_fir long
	short   nchans, chan;
	short		*temp_buf;
	long		grabsize, data_words;
  float   *flt_ptr_i[2], *flt_ptr_o[2], temp;
  float   *filter_ip[2], gain ; /* float version of buffer of each filter */
  float   maximum = -32768.0, minimum = 32768.0;
  FILE    *filter_from, *tempfile_fd;
  long    i, j, overflow[2] = {0L,0L}, undrflow[2] = {0L,0L}, sofar = 0L, update = 0L, ltemp;
	
	ULONG		data_bytes;
  
  gain = (float)0.9965;
	nchans = fh->nchans;
  
  /*
  For a full scale 1k tone with no 0 gain in the filter at 1k (e.g. ff.32k
  will 'slightly' overflow/underflow (0.03 dB) if the program gain = 1.0
  */
  
	data_bytes = fh->data_bytes;
	data_words = data_bytes/(2*nchans); // number of 16 bit integers in wav file
	if(verbose){
		printf("\nfir_sub1.c: Input  file: %s",inputname);
		printf("\nfir_sub1.c: Output file: %s\n",temp_fir_file);
		printf("\nnchans = %d",nchans);
		printf("\ndata_bytes = %d,  data_words = %d 16 bit words per channel\n\n",data_bytes,data_words);
  }

	
	
  /* Open the coefft File Needed */
  if ( (filter_from = fopen (fir_name, "rt")) == NULLPTR)
    error ("Can't open FILTER file:", fir_name);
		fseek(filter_from,0L,SEEK_SET);
  
	for ( chan = 0 ; chan < nchans ; chan++){
		if((filter_ip[chan] = malloc((SGMNT+2*MAX_HALF)*FLOATS)) == NULLPTR){
		  printf("fir_sub1.c: out of MEMORY for float INPUT area, chan = %d\n", chan);
			exit(1);
		}
	}

	if ( (tempfile_fd = fopen (temp_fir_file, "wb")) == NULLPTR)
    error ("Can't Create OUTPUT file:", temp_fir_file);
		fseek(tempfile_fd,0L,SEEK_SET);

	
	// total bytes = number of bytes in file
	
	write_wav_head32k(temp_fir_file, tempfile_fd, data_bytes, fh);
	
	/*--------------------------------------*/
	/* Initialise areas */
		zero_short_array(in, SGMNT, nchans); 
		zero_short_array(out, SGMNT, nchans);
		for ( chan = 0 ; chan < nchans ; chan++)
			zero_float_array(filter_ip[chan], (SGMNT+2*MAX_HALF));
		fir_size = rd_fir_file(filter_from, filter, gain);  /* read filter file and scale */
		half_fir = (fir_size-1)/2;
		fclose(filter_from); /* and close it */
		/*------------------------------------------------------------------------*/
		
		if((temp_buf = (short *)malloc(SHORTS*SGMNT*nchans)) == 0){ // use for wav file input & output buffer
			printf("fir_sub1.c: malloc error making temp_buf\n");
			exit(1);
		}
//		if(verbose) printf("\ntemp_buf made\n");

		grabsize = SGMNT; // 16384
		
		first_buf = TRUE;
		while( grabsize == SGMNT){
			//read file and demux if stereo      !!! IF DIFFERENT COEFFT FILES ARE NEEDED FOR THE TWO CHANNELS DO ONE CHANNEL AT A TIME
			if(nchans == 1){
//				if(verbose) printf("MONO\n");
				if((grabsize = fread(temp_buf,SHORTS,nchans*SGMNT,*reading_from)) < nchans*SGMNT){
					for(i=grabsize ; i < SGMNT ; i++) temp_buf[i] = 0;  // if needed fill up with zero's
				}// if(grabsize ...
				for(i = 0; i < SGMNT ; i++)	in[0][i] = temp_buf[i];
			} // end of if(nchans == 1)
			
			else if (nchans == 2) {
				if(verbose) printf("STEREO\n");
				// read stereo data into temp_buf
				grabsize = fread(temp_buf,SHORTS,nchans*SGMNT,*reading_from);
//				printf("1: grabsize = %d\n",grabsize);
				if(grabsize < nchans*SGMNT)	for(i=grabsize ; i < nchans*SGMNT ; i++) temp_buf[i] = 0;
				
				grabsize = grabsize/2;
				if(verbose) printf("grabsize after divide by 2 = %d\n",grabsize);
				// copy from temp_buf to in[][]
				for(i = 0; i < SGMNT ; i++){
					in[0][i] = temp_buf[2*i];
					in[1][i] = temp_buf[2*i+1];
				}
			} // else if(nchans == 2)
			// temp_buf finished use for reading
			
			/* enter main part of loop; repeat in SGMNT chunks until end of file   */
			
			for ( chan = 0 ; chan < nchans ; chan++ ){
				shrt_ptr[chan] = &out[chan][0];
				flt_ptr_i[chan] = filter_ip[chan]; /* initialise */
				convert_to_float(&in[chan][0], filter_ip[chan]+fir_size-1, grabsize);
				
				for (ix = 0; ix < grabsize; ix ++){
					temp = do_fir((flt_ptr_i[chan])++, filter, fir_size); /* FIR filtering */
					if (temp > SIGMAX){
						(overflow[chan])++; 
						if(temp > maximum) maximum = temp;
						temp = SIGMAX;
					}
					else if(temp < SIGMIN){ 
						(undrflow[chan])++; 
						if(temp < minimum) minimum = temp;
						temp = SIGMIN;
					}
					*(shrt_ptr[chan]++) = (short) temp;
				}
				if(first_buf) {
					if(chan == nchans-1){ // only write out grabsize-half_fir points  after chan 0 & 1 have been through
						// first_buf: chan 0 needs to wait for chan 1 so can put two chans together before writing to out
						if(grabsize > half_fir){
							if(nchans == 1){
//								printf("N1\n");
								if (fwrite (&out[0][half_fir], SHORTS, grabsize-half_fir, tempfile_fd) <= 0)
									error(" Error 1 in WRITING mono output data",NULLPTR);
							} else if (nchans == 2){
//								WRITE OUT STEREO_FILE
								for ( j = 0, i = half_fir ; i < grabsize - half_fir ; i++){
									temp_buf[j++] = out[0][i];
									temp_buf[j++] = out[1][i];
								} // for()
//								printf("Z1: 2*(grabsize-half_fir) SHORTS = %d\n",2*(grabsize-half_fir));
								if (fwrite (temp_buf, SHORTS, 2*(grabsize-half_fir), tempfile_fd) <= 0)
									error(" Error 1 in WRITING stereo output data",NULLPTR);
							}
						}else if (grabsize <= half_fir)	short_sig = TRUE; // sig length less than half fir length
						// i.e. less than delay caused by filtering
						first_buf = FALSE;
					}// if(chan == nchans-1)
					
				} else if (!first_buf){
					if(chan == nchans-1){// only write out points  after chan 0 & 1 
						if(nchans == 1){
//								printf("N1\n");
							if (fwrite (&out[0][0], SHORTS, grabsize, tempfile_fd) <= 0)
							error(" Error 2 in WRITING mono output data",NULLPTR);
						} else if (nchans == 2){
//								WRITE OUT STEREO_FILE
								for ( j=0, i = 0 ; i < grabsize  ; i++){
									temp_buf[j++] = out[0][i];
									temp_buf[j++] = out[1][i];
								} // for()
//								printf("Z2 2*grabsize SHORTS = %d\n",2*grabsize);
								if (fwrite (temp_buf, SHORTS, 2*grabsize, tempfile_fd) <= 0)
									error(" Error 2 in WRITING stereo output data",NULLPTR);
						}
					}
					
					/* shuffle data down through array ready for next loop */
				}// if(!first_buf)
				
				flt_ptr_i[chan] = filter_ip[chan]+grabsize;
				flt_ptr_o[chan] = filter_ip[chan];
				
				for (ix = 0; ix < fir_size-1; ix++){
					*((flt_ptr_o[chan])++) = *(flt_ptr_i[chan]);
					*((flt_ptr_i[chan])++) = 0.; /* safety to obliterate old data in wrong place */
				}
				sofar += grabsize;
				update += 1;
				
				/* don't print so often to the screen */
				//					if ((update & UPDATE) == UPDATE) printf("\r%7ld",sofar);fflush(stdout);
				
				if(verbose) printf("%7ld\n",sofar);fflush(stdout);
				
			} // for (chan = 0 ; chan < nchans ; chan++)
		} // while( grabsize == SGMNT*nchans )
		
			/* Signal is delayed by half_fir = half of (number of taps - 1)
			samples in the fir filter. Push delayed signal out of buffer
		by inputting enough zeroes */
		
		grabsize = half_fir; /* half number of taps */
		for ( chan = 0 ; chan < nchans ; chan++ ){
			shrt_ptr[chan] = &out[chan][0];
			
			flt_ptr_i[chan] = filter_ip[chan]; /* initialise */
			
			for (ix = 0; ix < grabsize; ix ++)  in[chan][ix] = 0;
			
			convert_to_float(&in[chan][0], filter_ip[chan]+fir_size-1, grabsize);
			
			for (ix = 0; ix < grabsize; ix ++){
				temp = do_fir((flt_ptr_i[chan])++, filter, fir_size); /* FIR filtering */
				if (temp > SIGMAX){
					(overflow[chan])++; 
					if(temp > maximum) maximum = temp;
					temp = SIGMAX;
				}
				else if(temp < SIGMIN){ 
					(undrflow[chan])++; 
					if(temp < minimum) minimum = temp;
					temp = SIGMIN;
				}
				*(shrt_ptr[chan]++) = (short) temp;
			}
			
			
			if(chan == nchans-1){
				if(short_sig){
					if(nchans == 1){
//								printf("N1\n");
						if (fwrite (&out[0][half_fir-data_words], SHORTS, data_words, tempfile_fd) <= 0)
							error(" Error 3 in WRITING mono output data",NULLPTR);
					} else if(nchans == 2){
//					WRITE OUT STEREO_FILE
						for ( j=0, i = half_fir-data_words ; i < data_words ; i++){
							temp_buf[j++] = out[0][i];
							temp_buf[j++] = out[1][i];
						} // for()
						printf("Z3 2*data_words SHORTS = %d\n",2*data_words);
						if ((ltemp = fwrite (temp_buf, SHORTS, 2*data_words, tempfile_fd)) <= 0)
							error(" Error 3 in WRITING stereo output data",NULLPTR);
						printf("Z3 ltemp (2*data_words) = %d\n",ltemp);
					}
				} else if(!short_sig){
					if(nchans == 1){
//						printf("N1\n");
						if (fwrite (&out[0][0], SHORTS, grabsize, tempfile_fd) <= 0)
						error(" Error 4 in WRITING mono output data",NULLPTR);
					} else if(nchans == 2){
//					WRITE OUT STEREO_FILE
						for ( j=0, i = 0 ; i < grabsize ; i++){
							temp_buf[j++] = out[0][i];
							temp_buf[j++] = out[1][i];
						} // for()
						printf("Z4 2*grabsize SHORTS = %d\n",2*grabsize);
						if (fwrite (temp_buf, SHORTS, 2*grabsize, tempfile_fd) <= 0)
							error(" Error 4 in WRITING stereo output data",NULLPTR);
					}
				}
			}// chan == nchans-1

		}// for (chan = 0 ; chan < nchans ; chan++)
//		if(verbose)printf("going to free temp_buf\n");
		free(temp_buf);
//		if(verbose)printf("temp_buf freed\n");
		/*  printf("\r%7ld 16-bit integers",sofar);*/
		fclose(*reading_from);
		fclose(tempfile_fd);
		for ( chan = 0 ; chan < nchans ; chan++)
			free(filter_ip[chan]);
		

//#if 0
    if(nchans == 1){
			if(overflow[0])
				printf("\nfir_sub (fir_sub1.c) :  %7ld overflows,  maximum %7.1f = %6.2f dB\n", overflow[0], maximum, 20*log10(maximum/SIGMAX));
			if(undrflow[0])
				printf("\nfir_sub (fir_sub1.c) :  %7ld underflows, minimum %7.1f = %6.2f dB\n", undrflow[0], minimum, 20*log10(minimum/SIGMIN));
		}

		if(nchans == 2){
			if(overflow[0] || overflow[1])
				printf("\nfir_sub (fir_sub1.c) :  %7ld %7ld overflows,  maximum %7.1f = %6.2f dB\n", overflow[0], overflow[1], maximum, 20*log10(maximum/SIGMAX));
		
			if(undrflow[0] || undrflow[1])
				printf("fir_sub (fir_sub1.c) \n:  %7ld %7ld underflows, minimum %7.1f = %6.2f dB\n", undrflow[0],undrflow[1], minimum, 20*log10(minimum/SIGMIN));
		}
//#endif
//printf("exit at end of fir_sub1.c\n");
}
