//  Dec 1st 2003

/* riff_subs.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h> /* file status: for finding out file length */
#include <sys/types.h>
#include "readfile.h"

// #define SWAP2(s) ((((s) >> 8) & 0xFF) | (((s) << 8) & 0xFF00))  /* needed for unix endian stuff */

// #define SWAP4(l) ((((l) >> 24) & 0xFF) | ((((l) & 0xFF0000) >> 8) & 0xFF00) \
// | (((l) & 0xFF00) << 8) | ((l) << 24))

#define TRUE 		1
#define FALSE		0

/* copy of structure declared in main prog */
extern struct formathead{
	USHORT w_format;    /* 16 bit number = 1 for PCM */
	USHORT nchans;      /* 1 or 2 channels */
	ULONG  clockHz;     /* sampling clock */
	ULONG  bytes_sec;   /* clockHz*nchans*bytes_sample */
	USHORT bytes;       /* 2=16 bit mono, 4 = 16 bit stereo */
	USHORT bits_sample; /* 16 for us */
	ULONG  data_bytes;  /* from DATA chunk */
};

extern short verbose;

/*------------------------------------------------------------------*/
void write_wav_head32k(char* filename, FILE *fd, ULONG data_bytes, struct formathead *fh)
{
	char chunk_name[5];
	ULONG total_bytes, chunk_bytes, bytes_sec;
	USHORT w_format=1, bytes_sample, bits_sample=16;
	

//	printf("write_wav_head: data_bytes = %ld\n",data_bytes);
		
	// filename has to have been OPENED in calling routine
	
	// write RIFF Chunk  12 bytes
	strcpy(chunk_name,"RIFF");
	fwrite(chunk_name,4,1,fd);
// total_bytes in WAV header ignores 1st 8 bytes in file ('RIFF' + 4 bytes for total length for package to follow)
	total_bytes = data_bytes + 36; // 4 for remainder of RIFF chunk, 24 for FORMAT, 8 for data header
	fwrite(&total_bytes,sizeof(ULONG),1,fd);
	strcpy(chunk_name,"WAVE");
	fwrite(chunk_name,4,1,fd);
	

//fh->clockHz is set at the new sample rate BEFORE CALLING write_wav_head32k
	// write FORMAT Chunk     24 bytes
	strcpy(chunk_name,"fmt ");
	fwrite(chunk_name,4,1,fd);
	chunk_bytes = 16; // 16 bytes follow after this in format header
	fwrite(&chunk_bytes,sizeof(ULONG),1,fd);
	fwrite(&fh->w_format,sizeof(USHORT),1,fd); // always 1 for PCM
	fwrite(&fh->nchans,sizeof(USHORT),1,fd);   // 1 - mono,  2- stereo  same as for input file
	fwrite(&fh->clockHz,sizeof(ULONG),1,fd);   // Sample rate in Hz  usually 32k, preset in calling program
	bytes_sec = fh->clockHz * fh->nchans * 2;         // bytes per second
	fwrite(&bytes_sec,sizeof(ULONG),1,fd);
	if(fh->nchans == 1) bytes_sample = 2;        // 16 bit mono
	else if(fh->nchans == 2) bytes_sample = 4;             // 16 bit stereo
	else { printf("nchans must equal 1 or 2\n"); exit(1);}
	fwrite(&bytes_sample,sizeof(USHORT),1,fd);
	fwrite(&bits_sample,sizeof(USHORT),1,fd);  // always 16 for us
	
	// write DATA chunk header, data to follow
	strcpy(chunk_name,"data");
	fwrite(chunk_name,4,1,fd);
	fwrite(&data_bytes,sizeof(ULONG),1,fd);  // ready for the data to be appended
}
/*------------------------------------------------------------------*/


int read_headers(char* filename, FILE **infd, ULONG *total_bytes, ULONG *data_chunk, struct formathead *fh)
{
// file filename is opened (infd) but not closed in this function
	long bytes_left, result;  // long so can be -ve
	ULONG chunk_bytes;
  char chunk_name[5];

  if ((*infd = fopen(filename,"rb")) == NULL) {
		fprintf(stderr,"\nmain: Can't open input file: %s\n\n",filename);
		exit(1);
	}
	result = fseek(*infd,0L,SEEK_SET); // set to start of file
	if(result){
		printf("\nread headers: fseek problem\n");
		exit(0);
	}
  read_riff_head(filename, total_bytes, *infd); /* open file, read 1st 12 bytes ("RIFF"  TOTAL_BYTES   "WAVE" */
	
  bytes_left = *total_bytes - 12;
  do {
    if(bytes_left < 8) { 
      printf("There are only %ld bytes left when trying to read an 8 byte header,  exiting\n",bytes_left);
      exit(0);
    }
    read_head(chunk_name, &chunk_bytes, *infd); /* read 1st 8 bytes of next chunk */
		
    bytes_left = bytes_left - 8 - (long)chunk_bytes;
		
		if(bytes_left < 0){
			printf("read_headers: bytes left less than zero (%ld)\n",bytes_left);
			exit(0);
		}
		
    if(!strcmp(chunk_name,"fmt ")){
      read_format(chunk_bytes, *infd, fh);
    }
		
    else if(!strcmp(chunk_name,"data")){
			fh->data_bytes = chunk_bytes;
			*data_chunk = chunk_bytes;
			// RETURN WITH file pointer at start of data
			return 0;
    } 

    else { /* anything not fmt or data */
      if(verbose) printf("%s chunk found\n",chunk_name);
      read_bits(chunk_bytes, *infd);  // throw away unwanted stuff
    }
		
//    if(verbose) printf("%ld bytes_left\n",bytes_left);
    if(bytes_left == 1) bytes_left = 0; 
		/* should only occur when used to force an even number of bytes */
  } while (bytes_left > 0);

  return 0;
}


/*--------------------------------------------------------------------*/

void read_format(ULONG chunk_bytes, FILE *infd, struct formathead *fh)
{
  char dummy;
  ULONG i;
	
  fread(&fh->w_format,1,2,infd);
  if(fh->w_format != 1){
    printf("Not a PCM wave, exiting ...\n\n");
    exit(0);
  }
	
  fread(&fh->nchans,1,2,infd);
//	printf("read_format: fh_nchans %d\n",fh->nchans);
  fread(&fh->clockHz,1,4,infd);
  fread(&fh->bytes_sec,1,4,infd);
  fread(&fh->bytes,1,2,infd);
  fread(&fh->bits_sample,1,2,infd);
  if(fh->bits_sample != 16){
    printf("We need 16 bit integers, this file contains %d bit integers\n",fh->bits_sample);
    printf("exiting\n");
    exit(0);
  }
	
  if(chunk_bytes > 16){
//    if(verbose) printf("%d excess bytes discarded from FORMAT chunk\n",chunk_bytes-16);
    for(i = 0 ; i < chunk_bytes-16 ; i++){
      fread(&dummy,1,1,infd);
    }
  }
}
/*------------------------------------------------------------------*/

void read_riff_head(char* filename, ULONG *total_bytes, FILE *infd)
{
  int retval;
  ULONG stat_file_len;
  struct stat file_stat;
  struct RiffHead{
    char riff[5];  /* 4 characters:  RIFF */
    ULONG cksize;/* length of file - 8 bytes */
    char wave[5];  /* 4 characters:  WAVE */
  } rh;

	retval = stat(filename,&file_stat);
	
  if(retval) {
    printf("stat of file retval = %d\n",retval);
    exit(0);
  }
  stat_file_len = (long)file_stat.st_size;
	
//  FILE ALREADY OPENED
	
  /* RIFF header */
  fread(&rh.riff,4,1,infd); /*read in first four bytes should be RIFF */
  rh.riff[4] = '\0';
  if(strcmp(rh.riff,"RIFF")){
    printf("%s is not a RIFF file\n\n",filename);
    exit(0);
  }
	
  fread(&rh.cksize,sizeof(ULONG),1,infd);/* read in 2nd 4 bytes (long with total_size-8) */
  *total_bytes = rh.cksize + 8;

	if(*total_bytes != stat_file_len){
    printf("file length from 'stat' of %s is %ld bytes\n",filename,stat_file_len);
    printf("file length read from the WAV file %s is %ld\n",filename,*total_bytes);
    printf("there is a problem with this file, take care\n");
		printf("ignoring header file length\n\n");
		*total_bytes = stat_file_len;
		rh.cksize = stat_file_len - 8;
  }
	
  fread(&rh.wave,4,1,infd); /* read in first four bytes should be WAVE */
  rh.wave[4] = '\0';
  if(strcmp(rh.wave,"WAVE")){
    printf("not a WAVE file\n");
    exit(0);
  }
}
/*--------------------*/
void read_head(char* chunk_name, ULONG *chunk_bytes, FILE *infd)
{
	
  fread(chunk_name,4,1,infd);
  chunk_name[4] = '\0';

  fread(chunk_bytes,sizeof(ULONG),1,infd); /* chunk_bytes already an address */
}
/*--------------------*/

void read_bits(ULONG chunk_bytes, FILE *infd)
  // throw away unwanted stuff
{
  ULONG i;
  char dummy;
  for(i = 0 ; i < chunk_bytes ; i++) {
    fread(&dummy,1,1,infd);
		//    if(verbose) printf("dummy byte:  %c\n",dummy);
  }
	//  if(verbose) printf("\n");
}
