// st_loudness3.c        stereo loudness
#include "readfile.h"

extern short verbose;

#define NPTS_MAX				2048         // MAX input buffer & fft size
#define NPTS_MAX2				(NPTS_MAX/2) // MAX spectrum output / expatt input size
#define MAX_THR				  50  // max number of values in a headphone correction file

// global
long tot_blocks;
double *ch_inst_sone[2]; // used for passing instant sones back to main where it is saved in ch_inst_sones[0 or 1]

// #define constant 0.06169745   // redefine constant for binaural inhibition case approx4/3 * old constant
//#define constant 0.06339136
//#define constant 0.063026
#define constant   0.06310


void st_loudness3(long channel_no, char* sigfile, char* bkgfile, double dBmax, double calib_scale, double ex_include, ULONG av_from, ULONG av_to,
                  short av_to_flag, FILE *sig_fd, FILE *bkg_fd,	short bkg_end, short monaural, short do_scale_sig, short do_scale_bkg,
                  char* Nfilename, short Nflag, short loud_type,double **ch1_inst_sploud, double **ch2_inst_sploud)
{
  short *svector(long nl, long nh);
		// st_loudness3 for stereo signals loud_type is 4 for stereo_sig , 5 for stereo_bkg
		short some_comps;
    short last2;   // flag for marking when input buffer is fully flushed
    short sig_end; // flag for marking when signal input file has finished being read
    long pts_read;
    int itemp;
    long npts = NPTS_MAX, npts2 = NPTS_MAX2;
    
    long block, comp, nlast1, last_ms = 10;   // , block1
    long i, j, ncomps, realft_n;
    long hndx, indx;
    
    double total_dB;
    double clock;
    double max_compdB, Hz_per_pt;
    double intensity,Abs_const,frqHz,frqkHz;
    double erb,p,g, E_No;
    
    //		constant = 0.04738978, 
    double alpha, Ethrq_dB, G_ex, EthrqG;
    double sones, K, Enoise, EthrN;
    double temp, temp1, temp2, const2;
    double G_dB, AAA, BBB;
    double EthrQ;
    double max1, max2, min1, min2, frame_ms;
    char Sfilename[2][4][255]={"dummy", "Left_Sig_spectrum", "Left_Bkg_spectrum", "Left_Both_spectrum",
      "dummy","Right_Sig_spectrum","Right_Bkg_spectrum","Right_Both_spectrum"};  // files that will be written
    short S_fd_open[2][4] = {FALSE, FALSE, FALSE, FALSE,FALSE, FALSE, FALSE, FALSE};
    FILE *S_fd[2][4]={NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL}; // file descriptors for Spectrum files  (channels 0 & 1)
    
    // in the excitiation pattern & loudness calculations we use
    // arrays in the Numerical Recipes tradition i.e. index starts at 1
    
    short *iwave1, *iwave2;
    long	*index;
    
    double *sp_loud; //*sp_loud1, *sp_loud2;   // left,  right  & stereo
    double *hanwin;
    double *compdB_1,*compdB_2,*compdB_3;
    double *rwave1, *rwave2, *rwave3;
    double *sig_int, *mask_int, *compfq;
    double *erbdB, *Esig, *Emask;
    
    // don't use frq_split[0]
    // ERB Numbers =             2.76,10.77,17.24,  12.14, 27.18,  39.00
    double frq_split[7] = {0.0,79.0,500.0,1234.0,2531.0,4031.0,14984.0}; // Hz
    
    //		char Nfile[256];
    FILE *N_fd = NULL, *fd[3]={NULL,NULL,NULL};
    //================= end of declarations  ============*/
    
    iwave1     = svector(1l,(long)npts); // shorts input buffer for 16 bit wave signal
    iwave2     = svector(1l,(long)npts); // shorts input buffer for 16 bit wave bkg
    index      = lvector(1l,6l);
    
    hanwin     = dvector(1l,(long)NPTS_MAX);
    sp_loud    = dvector(1l,(long)N_Es); // Number of 1/4 ERBS, N_Es = 156, ERB 1 to 39 incl   for writing both & reading left channel
//    sp_loud1   = dvector(1l,(long)N_Es); // Number of 1/4 ERBS, N_Es = 156, ERB 1 to 39 incl	for reading right channel
//    sp_loud2   = dvector(1l,(long)N_Es); // Number of 1/4 ERBS, N_Es = 156, ERB 1 to 39 incl  for writing stereo output
    compdB_1  = dvector(1l,(long)NPTS_MAX2);
    compdB_2  = dvector(1l,(long)NPTS_MAX2);
    compdB_3  = dvector(1l,(long)NPTS_MAX2);
    
    rwave1       = dvector(1l,(long)NPTS_MAX);
    rwave2       = dvector(1l,(long)NPTS_MAX);
    rwave3       = dvector(1l,(long)NPTS_MAX);
    erbdB        = dvector(1l,(long)NPTS_MAX);
    Esig         = dvector(1l,(long)N_ERBS);
    Emask        = dvector(1l,(long)N_ERBS);
    
    sig_int   = dvector(1l,(long)NPTS_MAX2);
    mask_int  = dvector(1l,(long)NPTS_MAX2);
    compfq    = dvector(1l,(long)NPTS_MAX2);
    
    
    
    clock = 32000.0;
    max1 =  -1.0e10;
    max2 = -999.0;
    min1 =  1.0e10;
    min2 =  999.0;
    
    
    // adjust dBmax to the scaled level for this channel
    dBmax = dBmax - 20.0*log10(calib_scale);
    if(verbose) {
      printf("\nchannel %1d calibration:\n",channel_no);
      printf("dBmax: a full scale 16-bit 1k Hz sinusoid would give %.1f dB\n",dBmax);
      printf("calib_scale %.3f   =   %.1f dB\n",calib_scale,20.*log10(calib_scale));
      printf("do_scale_sig  (0/1)  %1d,    do_scale_bkg (0/1) %1d\n",do_scale_sig, do_scale_bkg);
    }
    
    if(verbose) {
      printf("\nsig input file: %s\n",sigfile);
      if(!bkg_end) printf("bkg  input file: %s\n",bkgfile);
      else printf("Measuring loudness of signal file in quiet 2\n");
      printf("all spectral components more than %.1f dB below the max are discarded\n",ex_include);
    }
    
    
    Hz_per_pt = clock/(double)npts;   // 32000/2048 = 15.625 Hz spacing
    
    // Only use index = 1 to 6: components 1-6, 7-33, 34-79, 80-162, 163-258, 259-959
    for ( i = 1 ; i < 7 ; i++)
      index[i] = (long)(frq_split[i]/Hz_per_pt + 1.0);
    
    if(index[6] > NPTS_MAX2-1) index[6] = NPTS_MAX2-1;
    
    frame_ms = 1.0; // (32 points * 1000/clock ms) = 1 for 32k clock
    
    if(verbose){
      if(!monaural ) printf("\nBinaural presentation\n\n");
      else printf("\nMonaural presentation\n\n");
    }
    //---------------------------------------------------------------------------
    /*		
    * smallest frame rate is 32 points = 1 ms at 32k clock, 
    *      frame_ms = 32000.0/clock  ! 32 points
    *      frame_ms = 1.0 for 32k clock
    
      * FFT fixed at 2048 points for lowest freqs (64 ms buffer, 1 ms frame)
    *--------------------------------------------------------------------------*/
    
    // set up hanning window weights
    //	sum = 0.0; for calc dB loss of window
    for(i = 1 ; i <= npts ; i++){
      hanwin[i] = 0.5*(1.0 - cos((PI2*(i-1))/(double)(npts-1)));
      //			sum += hanwin[i] * hanwin[i];   
    }
    
    /*------ for a Hamming window use: -----------------------------
    hamwin[i] = (0.54-0.46 * cos((PI2*(i-1))/(double)npts));
    ---------------------------------------------------------------*/
    
    // initialise input wave buffers to zero
    for ( i = 1 ; i <= npts ; i++){
      iwave1[i] = 0;
      iwave2[i] = 0;
    }
    
    
    // read in 1 ms each time moving the rest of the buffer along by 1 ms
    block = -32;
    sig_end = last2 = FALSE; // bkg_end has already been set
    nlast1 = 0;
    
    // This is where the main loop returns for each block of data
    // last2 is only set when the 2048 point input buffer has been flushed
    //		while((!last2) && (block < (long)av_to))[
    while(!last2){
      block++; // a counter of the number of 32 data blocks read in - 32
      if(verbose) printf("\nblock %ld\n",block);
      
      // move buffer along by 32 samples
      for( i = 1 ; i <= npts-32 ; i++ ){
        iwave1[i] = iwave1[i+32];
        iwave2[i] = iwave2[i+32];
      }
      
      // read next 32 samples from the files or if finished reading file fill it with zeros
      // the end of either file finishes file input by setting sig_end = TRUE
      // at this point the 2048 point buffer will still be nearly full (64 *32 points)
      
      if((block > (long)av_to) && av_to_flag){
        //				printf("block %ld, av_to %ld    break:\n",block,av_to);
        sig_end = TRUE;
        break; // get out of while loop
      }
      
      if(!sig_end){
        // read from signal file
        if ((pts_read = (long)fread(&iwave1[1+npts-32],SHORTS,32,sig_fd)) != 32){
          if(ferror(sig_fd)){
            fprintf(stderr,"\nsig file %s read error (ferror)\n",sigfile);
            exit(1);
          }
          // check end of signal file
          if(feof(sig_fd)){
            if(verbose){
              printf("\nEnd of signal file %s (%ld points read)\n",sigfile,pts_read);
              printf("filling up with %ld zeros\n",32-pts_read);
            }
          }
          for (i=1+pts_read+npts-32; i <= npts ; i++) iwave1[i] = 0;
          sig_end = TRUE;
          if(verbose) printf("\nFinished reading signal file, ready to flush 64 by 32 points out\n");
        }	// if((pts_read.... for signal
      } // if(!sig_end)
      
      // for scaling just need to scale iwave1 from iwave1[1+npts-32] to iwave1[npts]  i.e. [i+npts-32] i=1 to 32 incl
      if(do_scale_sig) for( i = 1 ; i <= 32 ; i++) iwave1[i+npts-32] *= calib_scale;
      
      
      // read from bkg file
      if(!bkg_end){
        if ((pts_read = (long)fread(&iwave2[1+npts-32],SHORTS,32,bkg_fd)) != 32){
          // check end of bkg file
          if(ferror(sig_fd)){
            fprintf(stderr,"\nsig file %s read error\n",sigfile);
            exit(1);
          }
          // check end of signal file
          if(feof(sig_fd)){
            if(verbose){
              printf("\nEnd of bkg file %s  (%ld points read)\n",bkgfile,pts_read);
              printf("filling up with %ld zeros\n",32-pts_read);
            }
          }
          for (i=1+pts_read+npts-32; i <= npts ; i++) iwave2[i] = 0;
          bkg_end = TRUE;
          if(verbose){
            printf("\nFinished reading bkg file, emptying buffer (64-ms) \nthen");
            printf(" continuing with no background bkg (if necessary)\n");
          }
        }	// if((pts_read.... for bkg
        if(do_scale_bkg) for( i = 1 ; i <= 32 ; i++) iwave2[i+npts-32] *= calib_scale;
      }
      
      if(sig_end) { 
        // fill last 32 points of buffer with zeros instead of reading new values in
        for (i=1+npts-32; i <= npts ; i++) iwave1[i] = 0;
        nlast1++;
      }
      if(bkg_end){
        for (i=1+npts-32; i <= npts ; i++) iwave2[i] = 0;
      }
      // convert integer to real max scale +- 1 & window
      
      //------------------------------ start 2048 point 64-ms buffer --------------------------------*/
      // HANNING WINDOW ALL 2048 POINTS
      for( i = 1 ; i <= npts ; i++) {
        rwave1[i] = hanwin[i] * (double)iwave1[i]/32768.0;
        rwave2[i] = hanwin[i] * (double)iwave2[i]/32768.0;
        rwave3[i] = rwave1[i] + rwave2[i];
      }  // 201
      
      // this version is from 1st ed Num Rec, use correct realft() & four1()
      realft_n = npts2;
      
      realft(rwave1,realft_n,1);
      rwave1[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave1, npts, dBmax, 2048L);
      
      realft(rwave2,realft_n,1);
      rwave2[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave2, npts, dBmax, 2048L);
      
      realft(rwave3,realft_n,1);
      rwave3[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave3, npts, dBmax, 2048L);
      
      // copy index[1] points spectrum from 2048 window
      for ( i = 1 ; i <= index[1] ; i++){
        compdB_1[i] = rwave1[i];
        compdB_2[i] = rwave2[i];
        compdB_3[i] = rwave3[i];
      }
      
      // ---------------------------------------- 2048 done -----------------*/
      // extract 1024 point buffer into zero filled 2048 point buffer and window the 1024 points
      for ( i = 1 ; i <= npts ; i++ ){
        rwave1[i] = 0.0;
        rwave2[i] = 0.0;
        rwave3[i] = 0.0;
      }
      
      j = 0;
      // 2048 * 16/64 + 1 to 2048 * 48/64
      for ( indx = 513 ; indx <= 1536 ; indx++ ){
        j = j+1;
        hndx = (j-1)*2+1;
        rwave1[indx] = hanwin[hndx]*iwave1[indx]/32768.0;
        rwave2[indx] = hanwin[hndx]*iwave2[indx]/32768.0;
        rwave3[indx] = rwave1[indx] + rwave2[indx];
      }
      
      // now have rwave1() windowed & padded out to 2048 points ready for fft
      realft(rwave1,realft_n,1);
      rwave1[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave1, npts, dBmax, 1024L);
      
      realft(rwave2,realft_n,1);
      rwave2[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave2, npts, dBmax, 1024L);
      
      realft(rwave3,realft_n,1);
      rwave3[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave3, npts, dBmax, 1024L);
      
      // copy to compdB[] dB buffer
      for ( i = index[1]+1 ; i <= index[2] ; i++ ){
        compdB_1[i] = rwave1[i];
        compdB_2[i] = rwave2[i];
        compdB_3[i] = rwave3[i];
      }
      // ---------------------- 1024 done ---------------------------------
      // extract 512 point buffer into zero filled 2048 point buffer and window the 512 points
      for ( i = 1 ; i <= npts ; i++ ){
        rwave1[i] = 0.0;
        rwave2[i] = 0.0;
        rwave3[i] = 0.0;
      }
      
      j = 0;
      
      // 2048 * 24/64 + 1 to 2048 * 40/64
      for ( indx = 769 ; indx <= 1280 ; indx++ ){
        j = j+1;
        hndx = (j-1)*4+1;
        rwave1[indx] = hanwin[hndx]*iwave1[indx]/32768.0;
        rwave2[indx] = hanwin[hndx]*iwave2[indx]/32768.0;
        rwave3[indx] = rwave1[indx] + rwave2[indx];
      }
      
      // now have rwave1() windowed & padded out to 2048 points ready for fft
      realft(rwave1,realft_n,1);
      rwave1[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave1, npts, dBmax, 512L);
      
      realft(rwave2,realft_n,1);
      rwave2[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave2, npts, dBmax, 512L);
      
      realft(rwave3,realft_n,1);
      rwave3[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave3, npts, dBmax, 512L);
      
      // copy to compdB[] dB buffer
      for ( i = index[2]+1 ; i <= index[3] ; i++ ){
        compdB_1[i] = rwave1[i];
        compdB_2[i] = rwave2[i];
        compdB_3[i] = rwave3[i];
      }
      //--------------------------- 512 done ---------------------------*/
      // extract 256 point buffer into zero filled 2048 point buffer and window the 256 points
      for ( i = 1 ; i <= npts ; i++ ){
        rwave1[i] = 0.0;
        rwave2[i] = 0.0;
        rwave3[i] = 0.0;
      }
      
      j = 0;
      
      // 2048 * 28/64 + 1 to 2048 * 18/32
      for ( indx = 897 ; indx <= 1152 ; indx++ ){
        j = j+1;
        hndx = (j-1)*8+1;
        rwave1[indx] = hanwin[hndx]*iwave1[indx]/32768.0;
        rwave2[indx] = hanwin[hndx]*iwave2[indx]/32768.0;
        rwave3[indx] = rwave1[indx] + rwave2[indx];
      }
      
      // now have rwave1() windowed & padded out to 2048 points ready for fft
      realft(rwave1,realft_n,1);
      rwave1[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave1, npts, dBmax, 256L);
      
      realft(rwave2,realft_n,1);
      rwave2[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave2, npts, dBmax, 256L);
      
      realft(rwave3,realft_n,1);
      rwave3[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave3, npts, dBmax, 256L);
      
      // copy to compdB[] dB buffer
      for ( i = index[3]+1 ; i <= index[4] ; i++ ){
        compdB_1[i] = rwave1[i];
        compdB_2[i] = rwave2[i];
        compdB_3[i] = rwave3[i];
      }
      //--------------------------- 256 done ---------------------------*/
      // extract 128 point buffer into zero filled 2048 point buffer and window the 128 points
      for ( i = 1 ; i <= npts ; i++ ){
        rwave1[i] = 0.0;
        rwave2[i] = 0.0;
        rwave3[i] = 0.0;
      }
      
      j = 0;
      
      // 2048 * 30/64 + 1 to 2048 * 34/64
      for ( indx = 961 ; indx <= 1088 ; indx++ ){
        j = j+1;
        hndx = (j-1)*16+1;
        rwave1[indx] = hanwin[hndx]*iwave1[indx]/32768.0;
        rwave2[indx] = hanwin[hndx]*iwave2[indx]/32768.0;
        rwave3[indx] = rwave1[indx] + rwave2[indx];
      }
      
      // now have rwave1() windowed & padded out to 2048 points ready for fft
      realft(rwave1,realft_n,1);
      rwave1[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave1, npts, dBmax, 128L);
      
      realft(rwave2,realft_n,1);
      rwave2[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave2, npts, dBmax, 128L);
      
      realft(rwave3,realft_n,1);
      rwave3[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave3, npts, dBmax, 128L);
      
      // copy to compdB[] dB buffer
      for ( i = index[4]+1 ; i <= index[5] ; i++ ){
        compdB_1[i] = rwave1[i];
        compdB_2[i] = rwave2[i];
        compdB_3[i] = rwave3[i];
      }
      //--------------------------- 128 done ---------------------------*/
      // extract 64 point buffer into zero filled 2048 point buffer and window the 64 points
      for ( i = 1 ; i <= npts ; i++ ){
        rwave1[i] = 0.0;
        rwave2[i] = 0.0;
        rwave3[i] = 0.0;
      }
      
      j = 0;
      
      // 2048 * 31/64 + 1 to 2048 * 33/64
      for ( indx = 993 ; indx <= 1056 ; indx++ ){
        j = j+1;
        hndx = (j-1)*32+1;
        rwave1[indx] = hanwin[hndx]*iwave1[indx]/32768.0;
        rwave2[indx] = hanwin[hndx]*iwave2[indx]/32768.0;
        rwave3[indx] = rwave1[indx] + rwave2[indx];
      }
      
      // now have rwave1() windowed & padded out to 2048 points ready for fft
      realft(rwave1,realft_n,1);
      rwave1[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave1, npts, dBmax, 64L);
      
      realft(rwave2,realft_n,1);
      rwave2[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave2, npts, dBmax, 64L);
      
      realft(rwave3,realft_n,1);
      rwave3[2] = 0.0; // this is the NYQUIST freq output from realft
      scalefft(rwave3, npts, dBmax, 64L);
      
      for ( i = index[5]+1 ; i <= index[6] ; i++ ){
        compdB_1[i] = rwave1[i];
        compdB_2[i] = rwave2[i];
        compdB_3[i] = rwave3[i];
      } // 411
      for ( i = index[6]+1 ; i <= npts2 ; i++ ){ // all components above frq_split[6] Hz
        compdB_1[i] = -200.0;
        compdB_2[i] = -200.0;
        compdB_3[i] = -200.0;
      } // 421
      //--------------------------- 64 done ---------------------------*/
      
      // set up compfq[] with the component frequencies
      for ( i = 1 ; i <= npts2 ; i++)
        compfq[i] = (double)(i-1) * Hz_per_pt;
      
      // Now have a 1024 point spectrum in compdB_?[] & freqs in compfq() for ex pattern*/
      
      max_compdB = -200.0;
      // get max component level
      for ( i = 3 ; i <= npts2 ; i++){
        if(compdB_3[i] > max_compdB){
          max_compdB = compdB_3[i];
          temp = compfq[i];
        }
      }
      if(verbose)
        printf("Max spectral component is %7.2f dB at %.1f Hz\n",max_compdB,temp);
      
      total_dB = 0.0;
      some_comps = TRUE;
      if(max_compdB < -30.0){
        some_comps = FALSE;
        temp = max_compdB;
        //				if (verbose && block >= 0) printf("(1) block %ld: No components above -30 dB\n",block);
      }
      if(some_comps) {
      /* take out frequencies below 30 Hz &
      take out lower levels of spectrum & move everything along to save time
        calculating the excitation pattern, call number of components left ncomps */
        
        /* everything more than ex_include below the max value or below -30 dB is ditched,
        ex_include defaults to 60 dB, can be changed on the command line
        e.g. to 80 dB with -x 80 */
        
        j = 1; // incremented for each accepted value
        temp = max_compdB - ex_include; // temp is lowest level of component in dB 
        if(temp < -30.0) temp = -30.0;
        for( i = 3 ; i <= npts2 ; i++) {
          if((compdB_3[i] < temp)||(compdB_3[i] < -30.0)) continue;
          compfq[j] = compfq[i];
          compdB_1[j] = compdB_1[i];
          compdB_2[j] = compdB_2[i];
          compdB_3[j] = compdB_3[i];
          total_dB += pow(10.0,compdB_3[i]/10.0);
          j++;
        }// 451
        total_dB = 10.0*log10(total_dB);
        ncomps = j - 1;
        
        if(verbose){
          printf("ncomps used in excitation bit = %4d\n",ncomps);
          for( i = 1 ; i <= ncomps ; i++)
            printf("%4d freq %8.1f Hz %8.1f  %8.1f  %8.1f\n",    //  had dB where it should be Hz
            i,compfq[i],compdB_1[i],compdB_2[i],compdB_3[i]);
        }
      }else if(!some_comps)	{
        ncomps = 0;
      }
      
      if(ncomps < 1){
        if(verbose && block >= 0){
          printf("(2) block %ld: No components greater than -30 dB\n",block);
        }
        sones = 0.0;
        some_comps = FALSE;
      }
      if(verbose) printf("\t\t\t\t\t\tTotal dB = %.1f dB\n",total_dB);
      
      /*----------
      For each component in the 'signal+bkg' mix, convert the level in dB
      to linear intensity and store the values in array sig_int(),
      this is only used to calc the total level in the erb about each
      component. sig_int() this will later be used for the signal alone.
      this is for the 'signal + bkg'
      ------------ */
      
      for( i = 1 ; i <= ncomps ; i++){
        if(compdB_3[i] < -100.0) compdB_3[i] = -100.0; // might stop underflows
        sig_int[i] = pow(10.0,(double)(compdB_3[i]/10.0));
      }
      // 1st calc total level  in the normal roex(p) filter round each input component
      
      for( i = 1 ; i <= ncomps ; i++){
        intensity = 0.0;
        frqHz = compfq[i];
        frqkHz = frqHz/1000.0;
        erb = C1 * (C2 * frqkHz + 1.0);
        p = 4.0 * frqHz/erb;
        for( comp = 1 ; comp <= ncomps ; comp++){
          g = fabs((compfq[comp]-frqHz)/frqHz);
          if(g > 2.0) break;
          intensity += (1.0+p*g)*exp((double)(-p*g))*sig_int[comp];
        }
        erbdB[i] = 10.0 * log10((double)intensity);
      }
      
      
      
      // this is for the signal & bkg separately
      for( comp = 1 ; comp <= ncomps ; comp++){
        sig_int[comp] = pow(10.0,compdB_1[comp]/10.0);
        mask_int[comp] = pow(10.0,compdB_2[comp]/10.0);
      }
      
      calc_ex( Esig,compfq, sig_int,ncomps,erbdB);
      calc_ex(Emask,compfq,mask_int,ncomps,erbdB);
      
      Abs_const = 3.63;    
      sones = 0.0;
      // see from bottom of page 227 J.AES (1997) EthrqG is const
      EthrqG  = pow(10.0,(Abs_const/10.0));
      
      itemp = (int)(ESTART/ESTEP)-1;  // for 1.75/0.25 - 1 itemp = 6, used so sp_loud goes from E_No = 0.25 to 39
      for( i = 1 ; i <= itemp ; i++) sp_loud[i] = 0.0; // with E_No 0.25 to 1,5 having sp_loud[] = 0.0
      
      for ( i = 1 ; i <= N_ERBS ; i++ ){ // 621 (from original FORTRAN do loop label)
        sp_loud[(long)N_Es] = 0.0;
        sp_loud[i+itemp] = 0.0;
        E_No = ESTART + (i-1) * ESTEP; // maps sp_loud[i+temp] onto E_No e.g. i=1 --> sp_loud[7] for E_No = 1.75
        frqHz = erbtofq(E_No);
        frqkHz = frqHz/1000.0;
        
        // 500 Hz break between alpha=0.2 and alpha varying
        if(frqHz > 500.0){
          Ethrq_dB = Abs_const;
          G_ex = 1.0;
          G_dB = 0.0;
          alpha = 0.2;
          BBB = 2.0;
        }else if(frqHz <= 500.0){
          Ethrq_dB = ex_table(frqHz);
          G_ex = 1.0/pow(10.0,((Ethrq_dB - Abs_const)/10.0));
          G_dB = 10.0 * log10(G_ex);
          alpha = alpha100(G_dB);
          BBB = Btable(alpha);
        }
        
        EthrQ = pow(10.0,(Ethrq_dB/10.0));
        AAA = BBB * EthrqG;
        
        // ------ sp_loud calc for signal Esig[i] in bkg Emask[i] ---*/
        if(10.0*log10(Esig[i]) > Ethrq_dB-30.0){
          K = calc_K(frqkHz);
          Enoise  = Emask[i];
          // K*Enoise + EthrQ  =  K(Enoise + Eintbkg)
          EthrN = K*Enoise + EthrQ;
          
          if(Esig[i] >= EthrN){
            // signal excitation above threshold
            
            if(Esig[i]+Enoise <= 1.0e10){
              sp_loud[i+itemp] = constant * 
                ((pow(((Esig[i]+Enoise) * G_ex + AAA),  alpha) - pow(AAA,alpha))
                - (pow(((Enoise*(1.0+K)+EthrQ)*G_ex+AAA),alpha)
                - pow((EthrqG+AAA),alpha))  * pow((EthrN/Esig[i]),0.3));
            } else if(Esig[i]+Enoise > 1.0e10){
              const2 = constant/sqrt(1.04e6);
              sp_loud[i+itemp] = const2 * sqrt(Esig[i] + Enoise)
                - const2 * (sqrt((1.0+K)*Enoise+EthrQ)
                - pow((EthrqG+AAA),alpha) + pow(AAA,alpha))
                * pow((EthrN/Esig[i]),0.3);
            }
          } // if(Esig[i] >= EthrN)
          
          // signal excitation below threshold
          if(Esig[i] < EthrN) {
            if(Esig[i]+Enoise <= 1.0e10){
              sp_loud[i+itemp] = constant
                * (pow(((Esig[i]+Enoise) * G_ex + AAA),alpha)
                - pow((Enoise*G_ex+AAA),alpha));
              temp1 = pow((EthrQ * G_ex + AAA),alpha) - pow(AAA,alpha);
              temp2 = (pow(((EthrN+Enoise) * G_ex + AAA),alpha)
                - pow((Enoise*G_ex+AAA),alpha));
              sp_loud[i+itemp] = sp_loud[i+itemp] * (temp1/temp2)
                * pow(((2.0 * Esig[i])/(Esig[i] + EthrN)),1.5);
            }
            else if(Esig[i]+Enoise > 1.0e10){
              sp_loud[i+itemp] = constant*(sqrt(Esig[i]+Enoise)-sqrt(Enoise));
              temp1 = pow((EthrQ * G_ex + AAA),alpha) - pow(AAA,alpha);
              temp2 = sqrt(Enoise*(1.0+K)+EthrQ)- sqrt(Enoise);
              sp_loud[i+itemp] = sp_loud[i+itemp] * (temp1/temp2)
                * pow(((2.0 * Esig[i])/(Esig[i] + EthrN)),1.5);
            }
          } // if(Esig[i] < EthrN)
          
          if(sp_loud[i+itemp] < 0.0) 
            fprintf(stderr,"%7.1f Hz.  sp_loud less than 0  = %10.6f\n",frqHz,sp_loud[i+itemp]);
          
          sones = sones + sp_loud[i+itemp];
        }
      } // 621 end of sones calc for 1ms data scanned across N_ERBS
      
// Copy all 156 values of sp_loud[i] into ch1_inst_sploud[ERB index][block] for channel 1 (left)
// or ch2_inst_sploud[ERB index][block] for channel 2 (right)
      if(channel_no == 0){  // left channel
        if((block > 0) && (block < tot_blocks)){
          for ( i = 1 ; i <= N_Es ; i++) ch1_inst_sploud[i][block] = sp_loud[i];
//          printf("chan1 s block %4d:\n",block);
//          for ( i = 1 ; i <= N_Es ; i++) printf("%7.3f ",sp_loud[i]);
//          printf("\n\n");
        }
      }
      
      if(channel_no == 1){  // right channel
         if((block > 0) && (block < tot_blocks)){
          for ( i = 1 ; i <= N_Es ; i++) ch2_inst_sploud[i][block] = sp_loud[i];
//          printf("chan2 s block %4d:\n",block);
//          for ( i = 1 ; i <= N_Es ; i++) printf("%7.3f ",sp_loud[i]);
//          printf("\n\n");
        }
      }
      
      sones = sones * ESTEP;
      if(!monaural) sones = sones * 2.0; 
      
      if((block >= 0) && (block < tot_blocks)) ch_inst_sone[channel_no][block] = sones; // save the 'present' instant sones calculation
      // for channel channel_no and process in main after loudness() has been called      
      if(nlast1 >= 64) last2 = TRUE;
  } // while(!last2) go back for next ms of input waves until end of signal/bkg file
}
